__all__ = [
    'AbstractUpdatableModel', 'UpdatableModelMixin',
    'AbstractUpdate', 'UpdateMode',
    'UpdatableModelWrapper'
]

from .model import AbstractUpdatableModel, UpdatableModelMixin
from .update import AbstractUpdate, UpdateMode
from .wrapper import UpdatableModelWrapper



from typing import Generic, Iterable, TypeVar

from .model import AbstractModel

_Model = TypeVar('_Model', bound=AbstractModel)


class AbstractCompositeModel(AbstractModel, Generic[_Model]):
    def __init__(self, models: Iterable[_Model]):
        self._models = tuple(models)

    def __enter__(self):
        for m in self._models:
            m.__enter__()
        return self

    def __exit__(self, *exc):
        for m in self._models[::-1]:
            m.__exit__(*exc)



from typing import Iterable, Sequence, Tuple, Type, TypeVar

from tdm.abstract.datamodel import AbstractTreeDocumentContent
from tdm.datamodel import TalismanDocument

from tp_interfaces.abstract import AbstractCompositeModel, AbstractDocumentProcessor, ImmutableBaseModel

_DocumentContent = TypeVar('_DocumentContent', bound=AbstractTreeDocumentContent)
_Config = TypeVar('_Config', bound=ImmutableBaseModel)


class SequentialConfig(ImmutableBaseModel):
    def get_model_config(self, index: int) -> _Config:
        pass


class SequentialDocumentProcessor(
    AbstractCompositeModel[AbstractDocumentProcessor],
    AbstractDocumentProcessor[SequentialConfig, _DocumentContent]
):

    def __init__(self, processors: Iterable[AbstractDocumentProcessor]):
        AbstractCompositeModel[AbstractDocumentProcessor].__init__(self, processors)
        AbstractDocumentProcessor.__init__(self)

    def process_doc(self, document: TalismanDocument[_DocumentContent], config: _Config) -> TalismanDocument[_DocumentContent]:
        return self.process_docs([document], config)[0]

    def process_docs(self, documents: Sequence[TalismanDocument[_DocumentContent]], config: SequentialConfig) \
            -> Tuple[TalismanDocument[_DocumentContent], ...]:
        for _index, processor in enumerate(self._models):
            config_type = processor.config_type
            documents = processor.process_docs(documents, config_type())  # TODO implement correct configuration
        return documents

    @property
    def config_type(self) -> Type[SequentialConfig]:
        # TODO: generate runtime config
        return ImmutableBaseModel



from enum import Enum

from tp_interfaces.abstract.schema import ImmutableBaseModel


class UpdateMode(str, Enum):  # `str` is needed for correct JSON schema generation
    add = 'add'
    remove = 'remove'


class AbstractUpdate(ImmutableBaseModel):
    mode: UpdateMode



from typing import Callable, Dict, Generic, ItemsView, TypeVar

_Value = TypeVar('_Value')


class ModelTypeFactory(Dict[str, _Value], Generic[_Value]):
    def __init__(self, items: Dict[str, Callable[[], _Value]]) -> None:
        super().__init__(items)

    def __getitem__(self, k: str) -> _Value:
        return super().__getitem__(k)()

    def items(self) -> ItemsView[str, Callable[[], _Value]]:
        return super().items()



from abc import abstractmethod
from typing import Sequence, Type, TypeVar

from .model import AbstractModel

_Model = TypeVar('_Model', bound='AbstractConfigConstructableModel')


class AbstractConfigConstructableModel(AbstractModel):
    @classmethod
    @abstractmethod
    def from_config(cls: Type[_Model], config: dict) -> _Model:
        pass


class AbstractMultipleConfigConstructableModel(AbstractConfigConstructableModel):
    @classmethod
    def from_config(cls: Type[_Model], config: dict) -> _Model:
        return cls.from_configs([config])

    @classmethod
    @abstractmethod
    def from_configs(cls: Type[_Model], configs: Sequence[dict]) -> _Model:
        pass
